using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace MyDirect3DProject
{
	/// <summary>
	/// This is the main class of my Direct3D application
	/// </summary>
	public class MainClass : Form
	{
		/// <summary>
		/// The rendering device
		/// </summary>
		private Device device = null;
		private Texture tex;
		private int frame;	
		private const int numParticles = 500;
		
		public MainClass()
		{
			this.ClientSize = new System.Drawing.Size(640, 480);
			this.Text = "Direct3D Project";
		}
		
		public bool InitializeGraphics()
		{
			try {
				// Now let's setup the Direct3D stuff
				PresentParameters presentParams = new PresentParameters();
				presentParams.Windowed   = true;
				presentParams.SwapEffect = SwapEffect.Discard;
				presentParams.EnableAutoDepthStencil = true; // !!!				
				presentParams.AutoDepthStencilFormat = DepthFormat.D16; // !!!
				
				// Create the device
				device = new Device(0, DeviceType.Hardware, this, CreateFlags.SoftwareVertexProcessing, presentParams);
				
				// Setup the event handlers for the device
				device.DeviceLost     += new EventHandler(this.InvalidateDeviceObjects);
				device.DeviceReset    += new EventHandler(this.RestoreDeviceObjects);
				device.Disposing      += new EventHandler(this.DeleteDeviceObjects);
				device.DeviceResizing += new CancelEventHandler(this.EnvironmentResizing);
				
				tex = TextureLoader.FromFile(device, @"..\..\sprite.tga");
								
				return true;
			}
			catch (DirectXException)
			{
				return false;
			}
		}
		
		private VertexBuffer buildVertexBuffer()
		{
				VertexBuffer vb = new VertexBuffer(typeof(CustomVertex.PositionOnly),
													numParticles, device, Usage.WriteOnly, 0, Pool.Managed);
				CustomVertex.PositionOnly[] v = (CustomVertex.PositionOnly[]) vb.Lock(0, LockFlags.None);
				double t = 0.1*frame; // Bad: The apparent velocities of the particles will depend on the speed of the computer. 
				for(int i = 0; i < numParticles; i++)
				{
					double d = i/(double)numParticles;
					v[i] = new CustomVertex.PositionOnly((float)(Math.Sin(0.09*t+123.45*d)),
					                                     (float)(Math.Sin(0.23*t+456.78*d)),
					                                     (float)(Math.Sin(0.12*t+234.56*d)));
				}
				vb.Unlock();
				return vb;
		}
		
		protected virtual void InvalidateDeviceObjects(object sender, EventArgs e)
		{
		}
		
		protected virtual void RestoreDeviceObjects(object sender, EventArgs e)
		{
		}
		
		protected virtual void DeleteDeviceObjects(object sender, EventArgs e)
		{
		}
		
		protected virtual void EnvironmentResizing(object sender, CancelEventArgs e)
		{
		}
		
		/// <summary>
		/// This method moves the scene
		/// </summary>
		protected virtual void FrameMove()
		{
			// TODO : Frame movement
		}
		
		/// <summary>
		/// This method renders the scene
		/// </summary>
		protected virtual void Render()
		{
			if (device != null)
			{
				// Bad: Rebuilding the entire vertex buffer makes the graphics card wait for the CPU and vice versa.
				VertexBuffer vb = buildVertexBuffer();
				frame++;
				
				device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Color.Black, 1.0f, 0);
				device.BeginScene();

				device.RenderState.Lighting = false;
				
				device.VertexFormat = CustomVertex.PositionOnly.Format;
				device.SetStreamSource(0, vb, 0);
				device.SetTexture(0, tex);
				
				device.Transform.Projection = Matrix.PerspectiveFovLH(0.8f, Width/(float)Height, 0.1f, 50.0f);
				device.Transform.View = Matrix.LookAtLH(new Vector3(0.0f, 0.0f, -3.0f), new Vector3(), new Vector3(0.0f, 1.0f, 0.0f));
				device.Transform.World = Matrix.RotationY(0.2f);

				device.RenderState.PointSize = 0.1f;
				device.RenderState.PointScaleEnable = true;
				device.RenderState.PointScaleA = 0.00f;
				device.RenderState.PointScaleB = 0.00f;
				device.RenderState.PointScaleC = 1.00f;
				device.RenderState.PointSpriteEnable = true;
				
				device.RenderState.ZBufferWriteEnable = false;
				device.RenderState.AlphaBlendEnable = true;
				device.RenderState.SourceBlend = Blend.SourceAlpha;
				device.RenderState.DestinationBlend = Blend.One;
				
				device.DrawPrimitives(PrimitiveType.PointList, 0, numParticles);	
				
				device.EndScene();
				device.Present();
			}
		}
		
		/// <summary>
		/// Our main loop
		/// </summary>
		public void Run()
		{
			// While the form is still valid, render and process messages
			while (Created) {
				FrameMove();	
				Render();
				Application.DoEvents();
			}
		}
		
		protected override void OnPaint(PaintEventArgs e)
		{
			this.Render();
		}
		
		protected override void OnKeyPress(KeyPressEventArgs e)
		{
			base.OnKeyPress(e);
			if ((int)e.KeyChar == (int)System.Windows.Forms.Keys.Escape) {
				this.Close(); 
			}
		}
		
		/// <summary>
		/// The main entry point for the application
		/// </summary>
		static void Main()
		{
			using (MainClass mainClass = new MainClass()) {
				if (!mainClass.InitializeGraphics()) {
					MessageBox.Show("Error while initializing Direct3D");
					return;
				}
				mainClass.Show();
				mainClass.Run();
			}
		}
	}
}
